@@ -22,6 +22,9 @@ def default(val, d):
         return val
     return d() if callable(d) else d
 
+def divisible_by(numer, denom):
+    return (numer % denom) == 0
+
 def l2norm(t):
     return F.normalize(t, dim = -1)
 
@@ -39,11 +42,12 @@ def plain_cosine_sim_attention(
     v: TensorType['b', 'h', 'j', 'd'],
     mask: Optional[TensorType['b', 'j']] = None,
     attn_bias: Optional[TensorType['h', 'i', 'j']] = None,
-    scale = 8,
+    scale = 10,
     causal = False,
     l2norm_qk = True
 
 ) -> TensorType['b', 'h', 'i', 'd']:
+    assert not (causal and exists(mask)), 'mask should not be supplied if causality is needed'
 
     if l2norm_qk:
         q, k = map(l2norm, (q, k))
@@ -77,32 +81,41 @@ class FlashCosineSimAttention(Function):
         attn_bias,
         scale,
         causal,
-        q_block_size,
-        k_block_size,
-        tile_size,
-        backward_tile_size
+        backward_row_tile_size,
+        backward_col_tile_size,
+        backward_row_tiles,
+        backward_col_tiles
     ):
+        assert backward_col_tiles == 1
+        assert v.shape[-1] == 64, 'value dimension is fixed at 64 for now'
+
+        assert backward_row_tile_size == backward_col_tile_size
+
         batch, heads, seq, _, dim, device, dtype = *q.shape, v.shape[-1], q.device, q.dtype
 
-        mask = default(mask, lambda: torch.ones(q.shape[0], q.shape[2], device = q.device, dtype = torch.bool))
+        mask = default(mask, lambda: torch.ones(q.shape[0], 0, device = q.device, dtype = torch.bool))
 
-        o = torch.zeros((batch, heads, seq, dim), device = device, dtype = torch.float32)
-        l = torch.zeros((batch, heads, seq), device = device, dtype = torch.float32)
+        attn_bias = default(attn_bias, torch.empty(1, 0, 0, device = q.device, dtype = dtype))
 
-        attn_bias = default(attn_bias, torch.zeros(1, 0, 0, device = q.device, dtype = dtype))
-
-        forward(q, k, v, o, l, mask, attn_bias, scale, causal, q_block_size, k_block_size, tile_size)
-
-        o.div_(l[..., None].clamp(min = 1e-20))
+        o, l = forward(
+            q, k, v,
+            mask,
+            attn_bias,
+            scale,
+            causal
+        )
 
         ctx.save_for_backward(o, l, q, k, v, mask, attn_bias)
 
-        ctx.scale = scale
-        ctx.causal = causal
-        ctx.q_block_size = q_block_size
-        ctx.k_block_size = k_block_size
-        ctx.tile_size = tile_size
-        ctx.backward_tile_size = backward_tile_size
+        ctx.params = (
+            scale,
+            causal,
+            backward_row_tile_size,
+            backward_col_tile_size,
+            backward_row_tiles,
+            backward_col_tiles
+        )
+
         return o
 
     @staticmethod
@@ -111,23 +124,34 @@ class FlashCosineSimAttention(Function):
 
         batch, heads, src_seq, tgt_seq, device, dtype = *q.shape[:3], k.shape[2], q.device, q.dtype
 
-        scale = ctx.scale
-        causal = ctx.causal
-        q_block_size = ctx.q_block_size
-        k_block_size = ctx.k_block_size
-        tile_size = ctx.tile_size
-        backward_tile_size = ctx.backward_tile_size
-
-        dq, dk, dv = map(torch.zeros_like, (q, k, v))
+        (
+            scale,
+            causal,
+            backward_row_tile_size,
+            backward_col_tile_size,
+            backward_row_tiles,
+            backward_col_tiles
+        ) = ctx.params
 
         db = torch.zeros((heads, src_seq, tgt_seq), device = device, dtype = dtype) if attn_bias.requires_grad else torch.zeros((heads, 0, 0), device = device, dtype = dtype)
-        do_scaled = torch.zeros_like(l)
 
-        backward(do, o, l, q, k, v, dq, dk, dv, db, do_scaled, mask, attn_bias, scale, causal, q_block_size, k_block_size, backward_tile_size)
+        dq, dk, dv = backward(
+            do, o, l,
+            q, k, v,
+            db,
+            mask,
+            attn_bias,
+            scale,
+            causal,
+            backward_row_tile_size,
+            backward_col_tile_size,
+            backward_row_tiles,
+            backward_col_tiles
+        )
 
         db = db if attn_bias.requires_grad else None
 
-        return dq, dk, dv, None, db, None, None, None, None, None, None
+        return dq, dk, dv, None, db, None, None, None, None, None, None, None, None, None, None
 
 # wrapper function
 
@@ -137,16 +161,30 @@ def flash_cosine_sim_attention(
     v: TensorType['b', 'h', 'j', 'd'],
     mask: Optional[TensorType['b', 'j']] = None,
     attn_bias: Optional[TensorType['h', 'i', 'j']] = None,
-    scale = 8,
+    scale = 10,
     causal = False,
-    q_block_size = 64,
-    k_block_size = 64,
     l2norm_qk = True,
-    tile_size = 16,
-    backward_tile_size = 16
+    backward_row_tile_size = 16,
+    backward_col_tile_size = 16,
+    backward_row_tiles = 1,
+    backward_col_tiles = 1,
 ) -> TensorType['b', 'h', 'i', 'd']:
 
+    assert not (causal and exists(mask)), 'mask should not be supplied if causality is needed'
+
     if l2norm_qk:
         q, k = map(l2norm, (q, k))
 
-    return FlashCosineSimAttention.apply(q, k, v, mask, attn_bias, scale, causal, q_block_size, k_block_size, tile_size, backward_tile_size)
+    o = FlashCosineSimAttention.apply(
+        q, k, v,
+        mask,
+        attn_bias,
+        scale,
+        causal,
+        backward_row_tile_size,
+        backward_col_tile_size,
+        backward_row_tiles,
+        backward_col_tiles
+    )
+
+    return o
